-- Copyright(c)Chaos Group (www.chaosgroup.com). 
-- Feel free to distribute or modify this file. We would appreciate
-- if you keep the notice about where it originated from.

-- I will be glad if you drop me a note at vlado@chaosgroup.com
-- if you use it or modify it or just find it useful in any way.

-- SimCloth, revision 2
-- this is a simple script for simulating cloth movement
-- it's nothing like ClothReyes but might come handy
--
-- USAGE:
--    copy the script to \Scripts
--    copy the BMPs to \UI
--    evaluate this script (Run script... from the MAXScript menu),
--    customize a toolbar
--	  drag the SimCloth Icon from the ChaosGroup category

-- just to remind you - you use this script at your own risk.
-- I don't assume any responsibility for anything that you do with it,
-- or any damages that may be caused by its usage.


macroscript SimCloth category:"ChaosGroup" icon:#("simcloth",1) tooltip:"SimCloth by ChaosGroup"
(

-- UI globals
global cloth_main 
global cloth_floater
global ClothMesh = undefined
global gravity_node
global filter_deflectors


-- SimCloth global vars
global theMesh=undefined   -- the mesh to animate
global elastic=0.1         -- elasticity of cloth, should not be more than 0.5.
                           -- if the result trembles too much or loses form
  	                       -- altogether, decrease this
global airFriction=0.02    -- air friction, should not be more than 1.0
	                       -- larger values will result in more smooth motion
		  	 			   -- (less jelly-like)
global grav=[0,0,-0.2]     -- gravity
global deflect=#()         -- deflector nodes
global bondRange=2         -- something like valency, must be a positive integer, bigger values slow things down
                           -- the bigger the number, the more rigid the surface is
			        	   -- if this is higher, you should reduce elasticity
global slip=0.2            -- how "slippery" the cloth is when touching deflectors
global deflectRange=1.0    -- how "far" away from the deflector to start deflecting
global startTime=0         -- start time
global endTime=100         -- end time

----------------------------------------------------------
-- global vars


global arr=#() -- vertices
global vel=#() -- velocities of vertices

global ngb=#() -- for each vertex i, ngb[i] contains its neighbours

global bonds=#() -- bonds between vertices

-----------------------------------------------------------
-- functions

fn getVerts = (
	arr=#()
	vel=#()
	for i=1 to (getNumVerts theMesh) do (
		v=getVert theMesh i
		append arr v
		append vel [0,0,0]
	)
)

fn getNgb = (
	ngb=#()
	for i=1 to (getNumVerts theMesh) do append ngb #()
	for i=1 to (getNumFaces theMesh) do (
		face=getFace theMesh i
		for j=1 to 3 do (
			v1=0
			v2=0
			case j of (
				1: (v1=face.x; v2=face.y)
				2: (v1=face.y; v2=face.z)
				3: (v1=face.z; v2=face.x)
			)
			append ngb[v1] v2
			append ngb[v2] v1
		)
	)
)

-- adds a bond to the list, if it is not there already
fn addBond startIdx v1 v2 = (
	if (v1<v2) then (
		found=false
		for i=startIdx+1 to bonds.count do (
			if bonds[i][2]==v2 then (found=true; break)
		)
		if (not found) then (
			b=#(v1, v2, length ((getVert theMesh v2)-(getVert theMesh v1)))
			append bonds b
		)
	)
)

-- adds all bonds coinciding with a vertex
fn collectBonds startIdx s vn range = (
	if range>0 then (
		-- find all edges, coinciding with vertex vn
		for i=1 to ngb[vn].count do (
			addBond startIdx s ngb[vn][i]
			collectBonds startIdx s ngb[vn][i] (range-1)
		)
	)
)

fn getBonds = (
	bonds=#()
--	print "Listing bonds..."
	cloth_main.cloth_progress.color = color 180 0 0

	for i=1 to (getNumVerts theMesh) do (
		cloth_main.cloth_progress.value = (100.0*i/(getNumVerts theMesh))
		collectBonds bonds.count i i bondRange
--		format "% of % \n" i (getNumVerts theMesh)
	)
)

-- returns true if a vertex is fixed
fn fixedv v t =(
	(getVertSelection theMesh)[v]
)

-- cycle through the frames
fn anim meshSource = (

	cloth_main.cloth_progress.color = color 0 0 180

	with animate on (
		for t=startTime to endTime do (
			at time t (
			
				cloth_main.cloth_progress.value = (100.0 * t /endTime)
				
				
				cmesh=meshSource.mesh
				
				-- set the vertices
				-- setMesh m vertices:arr
				for i=1 to arr.count do (
					theMesh[4][1][i].controller.value=arr[i]
					--setProperty (theMesh[4][1]) ("vertex_"+(i as string)) arr[i]
				)
				update theMesh
						
				-- compute the new position of the vertices
				for i=1 to arr.count do (
					if (not (fixedv i t)) then arr[i]+=vel[i]
					else arr[i]=(getVert cmesh i)*meshSource.transform
					vel[i]*=1.0-airFriction
				)
				
				-- compute new velocities
				for i=1 to vel.count do (
					-- some gravity
					if (not fixedv i t) then vel[i]+=grav
					else vel[i]=[0,0,0]
				)
				
				-- forces along bonds
				for i=1 to bonds.count do (
					v1=bonds[i][1]
					v2=bonds[i][2]
					tlen=bonds[i][3]
					dir=arr[v2]-arr[v1]
					clen=length dir
					diff=(tlen-clen)/clen
					vel[v1]-=dir*diff*elastic
					vel[v2]+=dir*diff*elastic
				)
				
				-- deflect velocity
				for k=1 to deflect.count do (
					for i=1 to vel.count do (
						isect=intersectRay deflect[k] (ray arr[i] vel[i])
						if (isect!=undefined) then (
							t=(dot (isect.pos-arr[i]) vel[i])/(dot vel[i] vel[i])
							if (t<deflectRange) then (
								-- deflect the velocity
								nrm=normalize isect.dir
								tt=dot nrm vel[i]
								vel[i]=(vel[i]-nrm*tt)*slip -- vel[i]*t
							)						
						)
					)
				)
			)
		) -- for
	) -- animate
)

fn simCloth mesh = (
	meshSource=mesh
	theMesh=snapShot mesh
	theMesh.name = uniquename ("Cloth_"+mesh.name)
	animateVertex theMesh #all
	getVerts()
	getNgb()	
	getBonds()
	anim meshSource
	theMesh.transform=matrix3 1
	theMesh
)



--=============================================================
--CHAOS GROUP SIMCLOTH - USER INTERFACE
--=============================================================


if cloth_floater != undefined then closerolloutfloater cloth_floater 


rollout cloth_main "SimCloth Setup"
(
fn filter_mesh obj = superclassof obj == Geometryclass
fn filter_gravity obj = classof obj == Gravity
fn filter_wind obj = classof obj == Wind

pickbutton cloth_mesh ">Pick Cloth Source" width:200 filter:filter_mesh align:#left offset:[-8,0] across:3
button select_cloth "S" align:#right offset:[63,0] width:20 tooltip:"Select Cloth Source in Scene" 
checkbutton hide_cloth "H" align:#right checked:false offset:[9,0] width:20 tooltip:"Toggle Visibility - Show/Hide Cloth Source" highlightcolor:(color 255 200 200)

on hide_cloth changed state do
(
if ClothMesh != undefined then
(
if state then hide ClothMesh else unhide ClothMesh 
)
)

on select_cloth pressed do
(
if ClothMesh != undefined then 
(
max modify mode
select ClothMesh
)
)

group "Cloth Parameters"
(
spinner elastic_value "Elasticity  " range:[0,50,1] fieldwidth:40 across:2 align:#left
spinner bond_value "Bond Range" range:[1,100,2] fieldwidth:40 type:#integer align:#right
spinner slip_value "Slip Value" range:[1,100,20] fieldwidth:40 across:2 align:#left
spinner deflectrange_value "Deflect Range" range:[-100,100,1] fieldwidth:40 align:#right
)

on elastic_value changed val do elastic = val/100.0
on bond_value changed val do bondRange= val
on slip_value changed val do slip= val/100.0
on deflectrange_value changed val do deflectrange= val

group "Gravity"
(
radiobuttons use_gravity labels:#("User Values","Space Warp") 
spinner air_value "Air Friction" range:[0,1000,2] fieldwidth:50 align:#left across:2
spinner gravity_value "Gravity" range:[0,10,0.2] fieldwidth:50 align:#right
pickbutton gravity_vector ">Pick Gravity Vector" width:230 filter:filter_gravity align:#center tooltip:"Pick a Gravity Space Warp from Scene"
spinner gravity_x "X" range:[-1,1,0] across:3 align:#left fieldwidth:50
spinner gravity_y "Y" range:[-1,1,0] align:#center fieldwidth:50
spinner gravity_z "Z" range:[-1,1,-1] align:#right fieldwidth:50
)


on gravity_vector picked obj do
(
gravity_node = obj
gravity_vector.text = obj.name
grav = obj.dir * obj.strength
gravity_x.value = obj.dir.x
gravity_y.value = obj.dir.y
gravity_z.value = obj.dir.z
gravity_value.value = obj.strength
use_gravity.state= 2
gravity_x.enabled = false
gravity_y.enabled = false
gravity_z.enabled = false
gravity_value.enabled = false
)

on use_gravity changed state do
(
if state==1 then
(
gravity_x.enabled = true
gravity_y.enabled = true
gravity_z.enabled = true
gravity_value.enabled = true
)
else
(
gravity_x.enabled = false
gravity_y.enabled = false
gravity_z.enabled = false
gravity_value.enabled = false
)

)

--group "Wind"
--(
--pickbutton pick_wind ">Pick Wind" width:113 align:#left filter:filter_wind across:2 tooltip:"Pick a Wind Space Warp from Scene"
--button delete_wind "Remove Wind" width:113 align:#right 
--dropdownlist wind_list items:#() height:3
--)

on air_value changed val do airFriction= val/1000.0




fn filter_deflectors obj =
(
superclassof obj == Geometryclass and obj != ClothMesh and findItem deflect obj == 0
)

fn filter_deflector obj =
(
superclassof obj == Geometryclass and obj != ClothMesh 
)


group "Deflectors"
(
button select_deflectors ">Select Deflectors" width:115 align:#left enabled:false across:3
pickbutton pick_deflector ">Pick" width:55 offset:[20,0] align:#right enabled:false filter:filter_deflector
button delete_deflector "Remove" width:55 align:#right enabled:false 
listbox deflectors_list items:#() height:5
)

fn built_deflect_list d_l =
(
deflect = #()
for d = d_l.count to 1 by -1 do
(
try
(
d_o = execute ("$'"+d_l[d]+"'")
append deflect d_o
)
catch
(
deleteitem d_l d
)
)--end d loop
deflectors_list.items = d_l
)



on pick_deflector picked obj do
(
temp_array = deflectors_list.items  
if findItem temp_array obj.name == 0 then 
(
append temp_array obj.name
append deflect obj
deflectors_list.items = temp_array 
)
)

on delete_deflector pressed do
(
if deflectors_list.items.count > 0 then 
(
temp_array = deflectors_list.items
deleteItem temp_array deflectors_list.selection 
deflectors_list.items = temp_array 
built_deflect_list temp_array 
)
)

on select_deflectors pressed do
(
deflect_array = selectByName filter:filter_deflectors title:"Select SimCloth Deflectors"
temp_array = deflectors_list.items
if deflect_array != undefined then
(
for n in deflect_array do
(
if findItem temp_array n.name == 0 then append temp_array n.name
)
deflectors_list.items = temp_array
built_deflect_list temp_array 
)
)


group "Time Range"
(
radiobuttons use_time labels:#( "User Range","Scene Segment")
spinner start_time "Start Time" range:[0,10000,0] fieldwidth:50 align:#left across:2 type:#integer
spinner end_time "End Time" range:[0,10000,100] fieldwidth:50  align:#right type:#integer
)

on use_time changed state do
(
if state == 1 then
(
start_time.enabled = true
end_time.enabled = true
)
else
(
start_time.enabled = false
end_time.enabled = false
)
)

progressbar cloth_progress value:0 color:(color 0 0 180) width:240 height:10 align:#center 

button start_cloth "START CLOTH CALCULATION" width:240 height:30 align:#center enabled:false


on cloth_mesh picked obj do
(
ClothMesh = obj
cloth_mesh.text = obj.name
start_cloth.enabled = true
select_deflectors.enabled = true 
pick_deflector.enabled = true 
delete_deflector.enabled = true 
hide obj
hide_cloth.checked = true
)


group "Presets"
(
button save_presets "SAVE Presets" align:#left across:2 width:113
button load_presets "LOAD Presets" align:#right width:113
)

on save_presets pressed do
(
presets_file = getSaveFileName caption:"Save SimCloth Presets" types:"SimCloth Presets (*.scp)|*.scp"
if presets_file != undefined then
(
if ClothMesh != undefined then cm_name = ClothMesh.name else cm_name = "[none]"
out_file = createfile presets_file
format "[SimCloth Presets]\n" to:out_file
format "%\n" cm_name to:out_file

format "%\n" (elastic_value.value as string) to:out_file
format "%\n" (air_value.value as string) to:out_file
format "%\n" (bond_value.value as string) to:out_file
format "%\n" (slip_value.value as string) to:out_file
format "%\n" (deflectrange_value.value as string) to:out_file
format "%\n" (use_gravity.state as string) to:out_file
format "%\n" (gravity_x.value as string) to:out_file
format "%\n" (gravity_y.value as string) to:out_file
format "%\n" (gravity_z.value as string) to:out_file
format "%\n" (gravity_value.value as string) to:out_file

if gravity_node != undefined then
(
g_n = gravity_node.name
)
else
(
g_n = "[none]"
)

format "%\n" g_n to:out_file

format "%\n" (use_time.state as string) to:out_file
format "%\n" (start_time.value as string) to:out_file
format "%\n" (end_time.value as string) to:out_file

format "%\n" (deflectors_list.items.count as string) to:out_file
for d = 1 to deflectors_list.items.count do
(
format "%\n" deflectors_list.items[d] to:out_file
)
format "%\n" (hide_cloth.checked as string) to:out_file
format "[END OF FILE]\n" to:out_file
close out_file
)
)

on load_presets pressed do
(
presets_file = getOpenFileName caption:"Load SimCloth Presets" types:"SimCloth Presets (*.scp)|*.scp"
if presets_file != undefined then
(
in_file = openfile presets_file
str = readline in_file

str = readline in_file
try 
(
ClothMesh = execute ("$'"+str+"'")
cloth_mesh.text = str
start_cloth.enabled = true
select_deflectors.enabled = true 
pick_deflector.enabled = true 
delete_deflector.enabled = true 
)
catch
(
ClothMesh = undefined
)

try
(
str = readline in_file
elastic_value.value = execute str
str = readline in_file
air_value.value = execute str
str = readline in_file
bond_value.value = execute str
str = readline in_file
slip_value.value = execute str
str = readline in_file
deflectrange_value.value = execute str
str = readline in_file
use_gravity.state = execute str
str = readline in_file
gravity_x.value = execute str
str = readline in_file
gravity_y.value = execute str
str = readline in_file
gravity_z.value = execute str
str = readline in_file
gravity_value.value = execute str

if use_gravity.state ==1 then
(
gravity_x.enabled = true
gravity_y.enabled = true
gravity_z.enabled = true
gravity_value.enabled = true
)
else
(
gravity_x.enabled = false
gravity_y.enabled = false
gravity_z.enabled = false
gravity_value.enabled = false
)


g_n = readline in_file

str = readline in_file
use_time.state = execute str
str = readline in_file
start_time.value = execute str
str = readline in_file
end_time.value = execute str


if use_time.state == 1 then
(
start_time.enabled = true
end_time.enabled = true
)
else
(
start_time.enabled = false
end_time.enabled = false
)

str = readline in_file
d_c = execute str
d_l = #()
for d = 1 to d_c do
(
str = readline in_file
try
(
test_d = execute ("$'"+d_l+"'")
append d_l str
)catch()
)
str = readline in_file
hide_cloth.checked = execute str
if ClothMesh != undefined then 
(
if hide_cloth.checked then hide ClothMesh else unhide ClothMesh
)
close in_file
)
catch()

if d_l != undefined then
(
built_deflect_list d_l
)--end if d_l

try
(
gravity_node = execute ("$'"+g_n+"'")
gravity_vector.text = gravity_node.name
)
catch
(
gravity_node = undefined
gravity_vector.text = ">Pick Gravity Vector"
)

)--end if
)--end on



on start_cloth pressed do
(
if use_time.state == 1 then
(
startTime=start_time.value
endTime=end_time.value
)
else
(
startTime=animationrange.start as float / TicksPerFrame
endTime=animationrange.end as float / TicksPerFrame
)

if use_gravity.state == 2 and gravity_node != undefined then
(
grav = gravity_node.dir * gravity_node.strength
)
else
(
grav = (point3 gravity_x.value gravity_y.value gravity_z.value )* gravity_value.value
)

sliderTime = startTime
st = timestamp()
simCloth ClothMesh
et = timestamp()
print ((et-st)/1000.0)
cloth_progress.value = 0
)

)--end roll

rollout cloth_about "The Bulgarian Connection..."
(
label cloth_label01 "SimCloth Revision 2b"
label cloth_label03 "by CHAOS GROUP"
label cloth_label04 "http://www.chaosgroup.com"
label cloth_label05 "vlado@chaosgroup.com"
label cloth_label06 ""
label cloth_label07 "User Interface v1.1"
label cloth_label02 "13 January 2000"
label cloth_label08 "by Borislav 'Bobo' Petrov"
label cloth_label09 "http://gfxcentral.com/bobo"
)


cloth_floater = newrolloutfloater "SimCloth" 280 600 100 100
addrollout cloth_main cloth_floater 
addrollout cloth_about cloth_floater rolledup:true

format "Ready.\n"

)--end script

